-------------------------------------------------------------------------------
-- LISTENER by Tammya-MoonGuard (2016)
-------------------------------------------------------------------------------

local Main = ListenerAddon
local L = Main.Locale
local SharedMedia = LibStub("LibSharedMedia-3.0")
 
-- ordered list of chat history entries that should appear in the frame
Main.chat_buffer = {}
local g_chat_buffer_size = 300
local g_old_time = 60*10 -- time until messages are old

local g_has_unread_entries = false

-- when a new messages is added, this saves the time
-- if the user clicks on the frame at the moment a message is added, its ignore to prevent error
local g_clickblock = 0

local Me = {
	hl_texes = {};
	
	-- the current chatid
	chatid   = 0;
	showsay = true;
	showparty = true;
}

Main.Frame = Me

local function MixCol( a, b )
	
	local c = {
		a[1] + b[1];
		a[2] + b[2];
		a[3] + b[3];
	}
	
	local m = math.max( c[1], c[2] )
	m = math.max( m, c[3] )
	if m > 1.0 then
		c[1] = c[1] / m
		c[2] = c[2] / m
		c[3] = c[3] / m
	end
	
	c[1] = math.max( c[1], 0 )
	c[2] = math.max( c[2], 0 )
	c[3] = math.max( c[3], 0 )
	
	return c
end
   
function Main:HasUnreadEntries()
	return g_has_unread_entries
end
 
function Main:SetClickBlock()
	g_clickblock = GetTime()
end

function Main:OnChatboxScroll( delta )
	
	local reps = IsShiftKeyDown() and 5 or 1
	if delta > 0 then
		if IsAltKeyDown() then
			ListenerFrameChat:ScrollToTop()
		elseif IsControlKeyDown() then
			-- todo: increase font size
			Main:SetFontSize( Main.db.profile.frame.font.size + 1 )
		else
			for i = 1,reps do ListenerFrameChat:ScrollUp() end
		end
	else
		if IsAltKeyDown() then
			ListenerFrameChat:ScrollToBottom()
		elseif IsControlKeyDown() then
			-- todo: decrease font size
			Main:SetFontSize( Main.db.profile.frame.font.size - 1 )
		else
			for i = 1,reps do ListenerFrameChat:ScrollDown() end
		end
	end
	
	Main:UpdateUnreadHighlight()
end

function Main:OnChatboxHyperlinkClick( link, text, button )
	if GetTime() - g_clickblock < 0.4 then
		-- block clicks when scroll changes
		return
	end	
	 
	if ( strsub(link, 1, 6) == "player" ) then
		local namelink, isGMLink;
		if ( strsub(link, 7, 8) == "GM" ) then
			namelink = strsub(link, 10);
			isGMLink = true;
		else
			namelink = strsub(link, 8);
		end
		
		local name, lineid, chatType, chatTarget = strsplit(":", namelink);
		if IsShiftKeyDown() and button == "RightButton" then
			Main:RemovePlayer( name )
			return
		end
	end
	
	SetItemRef( link, text, button, DEFAULT_CHAT_FRAME );

end

-------------------------------------------------------------------------------
function Main:FlagUnreadEntries()
	ListenerFrameBarRead.tex:SetColorTexture( 0.95,0.90,0.13,0.8 )
	g_has_unread_entries = true
end

-------------------------------------------------------------------------------
function Main:UnflagUnreadEntries()
	ListenerFrameBarRead.tex:SetColorTexture( 0.95,0.90,0.13,0.2 )
	g_has_unread_entries = false
end

-------------------------------------------------------------------------------
function Main:TogglePlayerClicked()
	if UnitExists( "target" ) then 
		if UnitIsPlayer( "target" ) and UnitIsFriend( "target", "player" ) then
			if UnitGUID("target") == UnitGUID( "player" ) then return end
			
			self:TogglePlayer( UnitName( "target" ), true )
		end
	end
end

-------------------------------------------------------------------------------
function Main:ProbePlayer()
	local unit
	if UnitExists( "target" ) then 
		unit = "target"
	else
		unit = "mouseover"
	end
	
	local on = false
	
	if UnitGUID( unit ) == UnitGUID( 'player' ) then
		on = true
	elseif UnitIsPlayer( unit ) and UnitIsFriend( unit, "player" ) then
		local name = UnitName( unit )
		
		if name then
			on = self.player_list[name]
		end
	end
	
	if on then 
		ListenerFrameBarToggle.tex:SetColorTexture( 0.5,1,0.5,0.8 )
	else
		ListenerFrameBarToggle.tex:SetColorTexture( 0.5,1,0.5,0.2 )
	end
end

-------------------------------------------------------------------------------
function Main:ToggleSay()
	Me.showsay = not Me.showsay
	if Me.showsay then
		ListenerFrameBarShowSay.tex:SetColorTexture( 1,1,1,0.8 )
	else
		ListenerFrameBarShowSay.tex:SetColorTexture( 1,1,1,0.25 )
	end
	Main:RefreshChat()
	
	Main:Snoop_DoUpdate()
end

-------------------------------------------------------------------------------
function Main:ToggleParty()
	Me.showparty = not Me.showparty
	if Me.showparty then
		ListenerFrameBarShowParty.tex:SetColorTexture( 0.05,0.2,1,0.8 )
	else
		ListenerFrameBarShowParty.tex:SetColorTexture( 0.05,0.2,1,0.25 )
	end
	Main:RefreshChat()
	
	Main:Snoop_DoUpdate()
end

-------------------------------------------------------------------------------
function Main:ClipFramePosition()
	-- todo
end
 
-------------------------------------------------------------------------------
function Main:LoadFrameSettings()

	local left = Main.db.profile.frame.x or (GetScreenWidth()/2)
	local bottom  = Main.db.profile.frame.y or (GetScreenHeight()/2)
	
	left   = math.max( left, 0 )
	bottom = math.max( bottom, 0 )
	left   = math.min( left, GetScreenWidth() - 32 )
	bottom = math.min( bottom, GetScreenHeight() - 32 )
	
	if Main.db.profile.frame.width < 50 then
		Main.db.profile.frame.width = 50
	end
	
	if Main.db.profile.frame.height < 50 then
		Main.db.profile.frame.height = 50
	end
		
	ListenerFrame:SetPoint( "BOTTOMLEFT", left, bottom )
	ListenerFrame:SetSize( Main.db.profile.frame.width, Main.db.profile.frame.height )
	
	ListenerFrame:SetBackdropColor( Main.db.profile.frame.bg.r,
	                                Main.db.profile.frame.bg.g,
                                    Main.db.profile.frame.bg.b,
	                                Main.db.profile.frame.bg.a )
									
	ListenerFrame:SetBackdropBorderColor( Main.db.profile.frame.edge.r,
	                                      Main.db.profile.frame.edge.g,
                                          Main.db.profile.frame.edge.b,
	                                      Main.db.profile.frame.edge.a )
	
	Main:LoadChatFont()
	
	if Main.db.profile.frame.hidden then
		ListenerFrame:Hide()
	else
		ListenerFrame:Show()
	end
end

-------------------------------------------------------------------------------
function Main:LoadChatFont()
	local outline = nil
	if Main.db.profile.frame.font.outline == 2 then
		outline = "OUTLINE"
	elseif Main.db.profile.frame.font.outline == 3 then
		outline = "THICKOUTLINE"
	end
	local font = SharedMedia:Fetch( "font", Main.db.profile.frame.font.face )
	ListenerFrameChat:SetFont( font, Main.db.profile.frame.font.size, outline )
	
	if Main.db.profile.frame.font.shadow then 
		ListenerFrameChat:SetShadowColor( 0, 0, 0, 0.8 )
		ListenerFrameChat:SetShadowOffset( 1,-1 )
	else
		
		ListenerFrameChat:SetShadowColor( 0, 0, 0, 0 )
	end
end

-------------------------------------------------------------------------------
function Main:ToggleFrame()
	self.db.profile.frame.hidden = not self.db.profile.frame.hidden
	if self.db.profile.frame.hidden then
		ListenerFrame:Hide()
	else
		ListenerFrame:Show()
	end
end

-------------------------------------------------------------------------------
function Main:HideFrame( dontsave )
	ListenerFrame:Hide()
	if not dontsave then
		self.db.profile.frame.hidden = true
	end
end

-------------------------------------------------------------------------------
function Main:ShowFrame( dontsave )
	ListenerFrame:Show()
	if not dontsave then
		self.db.profile.frame.hidden = false
	end
end

-------------------------------------------------------------------------------
function Main:SaveFramePosition()
	self.db.profile.frame.x = ListenerFrame:GetLeft()
	self.db.profile.frame.y = ListenerFrame:GetBottom()
	self.db.profile.frame.width = ListenerFrame:GetWidth()
	self.db.profile.frame.height = ListenerFrame:GetHeight()
end

-------------------------------------------------------------------------------
local SAY_EVENTS = { SAY=true, EMOTE=true, TEXT_EMOTE=true, YELL=true }

-------------------------------------------------------------------------------
local function EntryFilter( entry )
	if SAY_EVENTS[entry.e] then
		return Me.showsay
	elseif Me.showparty then
		return true
	end
end

local function GetEntryColor( e )
	
	local color = Main.db.profile.colors[e.e]
	 
	if e.p then
		-- player message
		color = Main.db.profile.colors[ "P_" .. e.e ]
	end 
	
	return color
end

local g_msgtypes = { 
	SAY=1, PARTY=1, PARTY_LEADER=1, RAID=1, RAID_LEADER=1, RAID_WARNING=1, YELL=1;
	EMOTE=2;
	TEXT_EMOTE=3;
}	

local g_m1_prefix = {
	PARTY        = "[P] ";
	PARTY_LEADER = "[PL] ";
	RAID         = "[R] ";
	RAID_LEADER  = "[RL] ";
	RAID_WARNING = "[RW] ";
}

function Main:FormatChatMessage( e )
	local msgtype = g_msgtypes[e.e]
	
	-- get icon and name
	local name, icon, color --
	
	local stamp = ""
	if Main.db.profile.frame.timestamps then
		stamp = date( "[%H:%M] ", e.t )
		stamp = "|cff808080" .. stamp .. "|r"
	end
	
	if not e.p or Main.db.profile.frame.playername then 
		name, icon, color = self:GetICName( e.s, e.g )
		
		if icon then
			icon = "|TInterface\\Icons\\" .. icon .. ":0|t "
		else
			icon = ""
		end
		
		if color then
			name = "|c" .. color .. name .. "|r"
		end
		
		name = "|Hplayer:" .. e.s .. "|h" .. name .. "|h" 
	end
	 
	if msgtype == 1 then
	   
		local prefix = g_m1_prefix[e.e] or ""
		prefix = stamp .. prefix
		
		if e.p and not Main.db.profile.frame.playername then
			return string.format( "%s>> %s", prefix, e.m )
		else
			return string.format( "%s%s%s: %s", prefix, icon, name, e.m )
		end
	elseif msgtype == 2 then
		if e.p and not Main.db.profile.frame.playername  then
			return string.format( "%s** %s", stamp, e.m )
		else
			return string.format( "%s%s%s %s", stamp, icon, name, e.m )
		end
	elseif msgtype == 3 then
	
		if e.p and not Main.db.profile.frame.playername  then 
			return string.format( "%s** %s", stamp, e.m )
		else
			local msg = e.m:gsub( e.s, name )
			return string.format( "%s%s%s", stamp, icon, msg )
		end
	
	end
	return ""
	
end

-------------------------------------------------------------------------------
function Main:AddMessage( e, beep )
	if not EntryFilter( e ) then return end
	
	Me.chatid = Me.chatid + 1
	
	table.insert( self.chat_buffer, { e = e, c = Me.chatid } ) 
	while #self.chat_buffer > g_chat_buffer_size do 
		table.remove( self.chat_buffer, 1 )
	end
	
	if not e.r and not e.p then
		if beep and Main.db.profile.sound.msg and ListenerFrame:IsShown() then
			Main:PlayMessageBeep() 
		end
		self:FlagUnreadEntries()
	end
	
	local color = GetEntryColor( e )
	ListenerFrameChat:AddMessage( self:FormatChatMessage( e ), 
	                              color[1], color[2], color[3], Me.chatid )
end
--[[
-------------------------------------------------------------------------------
function Main:UpdateMessageColor( e )
	if not e.c or (e.c < Me.chatid - g_chat_buffer_size) then
		return -- out of range? this should be an error
	end
	
	local offset = math.max( Me.chatid - g_chat_buffer_size, 0 )
	local color = GetEntryColor( e ) 
	ListenerFrameChat:UpdateColorByID( e.c - offset, color[1], color[2], color[3] )
end]]

-------------------------------------------------------------------------------
function Main:CheckUnread()
	self:UnflagUnreadEntries()
	g_has_unread_entries = false
	for k,v in pairs( self.unread_entries ) do
		if EntryFilter( v ) and (self.player_list[ v.s ] or guid == UnitGUID("player")) then
			self:FlagUnreadEntries()
		end
	end
end

-------------------------------------------------------------------------------
function Main:UpdateUnreadHighlight()
	local regions = {}

	-- create a list of message regions
	for k,v in pairs( { ListenerFrameChat.FontStringContainer:GetRegions() } ) do
		
		if v:GetObjectType() == "FontString" and v:IsShown() then
			v:SetNonSpaceWrap( true ) -- a nice hack for nonspacewrap text
			table.insert( regions, v )
		end
	end
	
	-- sort by Y
	table.sort( regions, function( a, b ) 
		return a:GetTop() < b:GetTop()
	end)
	
	local offset = math.max( Me.chatid - g_chat_buffer_size, 0 )
	local chat_index = Me.chatid - ListenerFrameChat:GetScrollOffset() - offset
	local count = 0
	
	local top_edge = ListenerFrameChat:GetTop() + 1 -- that one pixel
	
	for k,v in ipairs( regions ) do
		local e = Main.chat_buffer[chat_index]
		
		if not e then break end
		e = e.e
	
		if not e.r and not e.p and v:GetBottom() < top_edge then
			
			-- unread message, highlight
			count = count + 1
			if not Me.hl_texes[count] then
				Me.hl_texes[count] = ListenerFrameChat:CreateTexture() 
			end 
			local tex = Me.hl_texes[count]
			
			tex:ClearAllPoints()
			tex:SetPoint( "LEFT", v, "LEFT", 0, 0 )
			tex:SetPoint( "RIGHT", v, "RIGHT", 0, 0 )
			tex:SetPoint( "BOTTOM", v, "BOTTOM", 0, 0 )
			
			local clip = math.max( v:GetTop() - top_edge, 0 )
			
			tex:SetPoint( "TOP", v, "TOP", 0, -clip )
			
			if Main.db.profile.colors.highlight_add then
				tex:SetBlendMode( "ADD" )
			else
				tex:SetBlendMode( "BLEND" )
			end
			tex:SetColorTexture( unpack( Main.db.profile.colors.highlight ) )
			tex:Show()
		end

		
		chat_index = chat_index - 1
	end
	
	for i = count+1, #Me.hl_texes do
		Me.hl_texes[i]:Hide()
	end
	 
end

-------------------------------------------------------------------------------
function Main:ClearChatBuffer()  
	self.chat_buffer = {}
	Me.chatid = 0
end

-------------------------------------------------------------------------------
function Main:RefreshChat()
	ListenerFrameChat:Clear()
	self:ClearChatBuffer()
	self:UnflagUnreadEntries()
	local entries = {}
	
	for playername,p in pairs(self.player_list) do
		local history = self.chat_history[playername] or {}
		for k2, v2 in pairs( history ) do
			if EntryFilter( v2 ) then
				table.insert( entries, v2 )
			end
		end
	end
	
	-- our own text
	local history = self.chat_history[UnitName("player")] or {}
	for k2, v2 in pairs( history ) do
		if EntryFilter( v2 ) then
			table.insert( entries, v2 )
		end
	end
	
	-- sort by lineid
	table.sort( entries, function( a, b )
		return a.id < b.id
	end)
	
	-- crop and populate
	local start = #entries - g_chat_buffer_size
	if start < 1 then start = 1 end
	for i = start,#entries do
		self:AddMessage( entries[i] )
	end 
	
	Main:UpdateUnreadHighlight()
end

-------------------------------------------------------------------------------
function Main:SetChatFont( font )
	self.db.profile.frame.font.face = font
	self:LoadChatFont()
end

-------------------------------------------------------------------------------
function Main:SetChatOutline( value )
	self.db.profile.frame.font.outline = value
	self:LoadChatFont()
end

-------------------------------------------------------------------------------
function Main:SetFontSize( size )
	size = math.max( size, 6 )
	size = math.min( size, 24 )
	self.db.profile.frame.font.size = size
	self:LoadChatFont()
end

-------------------------------------------------------------------------------
function Main:Frame_SetShadow( val )
	self.db.profile.frame.font.shadow = val
	self:LoadChatFont()
end	

-------------------------------------------------------------------------------
function Main:Frame_SetBGColor( r, g, b, a )
	self.db.profile.frame.bg.r = r
	self.db.profile.frame.bg.g = g
	self.db.profile.frame.bg.b = b
	self.db.profile.frame.bg.a = a
	self:LoadFrameSettings()
end

-------------------------------------------------------------------------------
function Main:Frame_SetEdgeColor( r, g, b, a )
	self.db.profile.frame.edge.r = r
	self.db.profile.frame.edge.g = g
	self.db.profile.frame.edge.b = b
	self.db.profile.frame.edge.a = a
	self:LoadFrameSettings()
end

function Main:Frame_SetTimestamps( stamps )
	self.db.profile.frame.timestamps = stamps
	Main:RefreshChat()
end

function Main:Frame_SetPlayerName( val )
	self.db.profile.frame.playername = val
	Main:RefreshChat()
end

-------------------------------------------------------------------------------
-- FRAME SCRIPTS
-------------------------------------------------------------------------------

function Main.Frame_OnEnter( self )
	
	Main.Frame_UpdateResizeShow()
end

-------------------------------------------------------------------------------
function Main.Frame_OnLeave( self )
	
	Main.Frame_UpdateResizeShow()
end

-------------------------------------------------------------------------------
function Main.Frame_OnMouseDown( self, button )
	if button == "LeftButton" then
		ListenerFrame.dragging = true
		ListenerFrame:StartMoving() 
	end
end

-------------------------------------------------------------------------------
function Main.Frame_OnMouseUp( self )
	
	Main.Frame_StopDragging()

end

function Main.Frame_StopDragging()
	if ListenerFrame.dragging then
		ListenerFrame.dragging = false
		ListenerFrame:StopMovingOrSizing()
		ListenerAddon:ClipFramePosition()
		ListenerAddon:SaveFramePosition()
	end
end

-------------------------------------------------------------------------------
function Main.FrameResizeButton_OnMouseDown( self, button )
	
	self:SetButtonState( "PUSHED", true ); 
	self:GetHighlightTexture():Hide();
	ListenerFrame:StartSizing( "BOTTOMRIGHT" );
	ListenerFrame.doingSizing = true
end

-------------------------------------------------------------------------------
function Main.FrameResizeButton_OnMouseUp( self, button )

	self:SetButtonState( "NORMAL", false );
	self:GetHighlightTexture():Show();
	ListenerFrame:StopMovingOrSizing();
	ListenerAddon:ClipFramePosition()
	ListenerAddon:SaveFramePosition()
	ListenerFrame.doingSizing = false
	
	Main.Frame_UpdateResizeShow()
end

-------------------------------------------------------------------------------
function Main.FrameResizeButton_OnLeave( self )
	
	-- hide ourselves, but we're shown again right after
	Main.Frame_UpdateResizeShow()
end

function Main.Frame_UpdateResizeShow()
	if ListenerFrame:IsMouseOver() and IsShiftKeyDown() or ListenerFrame.doingSizing then
		ListenerFrameResizeButton:Show()
	else
		ListenerFrameResizeButton:Hide()
	end
end

